/*===================================================================*/
/*                                                                   */
/*  PSX_Spu.c : Sound routine for PSX                                */
/*              based on  Dodger of CREATURE sample and              */
/*              BERO's mp3player sample                              */
/*                                                                   */
/*  2000/06/07  Racoon                                               */
/*                                                                   */
/*===================================================================*/

typedef unsigned char u_char;
typedef unsigned short u_short;
typedef unsigned long u_long;

/*===================================================================*/

#define i_reg  (volatile unsigned short*)0x1F801070
#define i_mask (volatile unsigned short*)0x1F801074
#define spu_mvol_l (volatile unsigned short*)0x1F801D80 
#define spu_mvol_r (volatile unsigned short*)0x1F801D82 
#define sp0 (volatile unsigned short*)0x1F801DAA // flag
#define sp1 (volatile unsigned short*)0x1F801DAE // status 
#define spu_keyON_1 (volatile unsigned short*)0x1F801D88 
#define spu_keyON_2 (volatile unsigned short*)0x1F801D8A
#define spu_keyOFF_1 (volatile unsigned short*)0x1F801D8C
#define spu_keyOFF_2 (volatile unsigned short*)0x1F801D8E
#define cd_mvol_l (volatile unsigned short*)0x1F801DB0 
#define cd_mvol_r (volatile unsigned short*)0x1F801DB2 
#define voice_base (volatile unsigned short*)0x1F801C00
#define spu_sb_addr (volatile unsigned short*)0x1F801DA6
#define spu_delay (volatile unsigned short*)0x1F801014 
#define spu_u_1 (volatile unsigned short*)0x1F801D84 // reverberation depth left
#define spu_u_2 (volatile unsigned short*)0x1F801D86 // reverberation depth right
#define spu_u_3 (volatile unsigned short*)0x1F801D98 
#define spu_u_4 (volatile unsigned short*)0x1F801D9A 
#define spu_u_5 (volatile unsigned short*)0x1F801DB4 
#define spu_u_6 (volatile unsigned short*)0x1F801DB6 
#define spu_u_7 (volatile unsigned short*)0x1F801D90
#define spu_u_8 (volatile unsigned short*)0x1F801D92
#define spu_u_9 (volatile unsigned short*)0x1F801D94 
#define spu_u_A (volatile unsigned short*)0x1F801D96
#define spu_u_B (volatile unsigned short*)0x1F801DAC
#define spu_u_D (volatile unsigned short*)0x1F801DA8
#define spu_u_E (volatile unsigned short*)0x1F801DA2
#define spu_dma_1 (volatile long*)0x1F8010C0
#define spu_dma_2 (volatile long*)0x1F8010C4
#define spu_dma_3 (volatile long*)0x1F8010C8
#define d_pcr (volatile long*)0x1F8010F0
#define d_icr (volatile long*)0x1F8010F4
#define t1_mode (volatile long*)0x1F801114

/*===================================================================*/
/*                                                                   */
/*                          SPU_Enable()                             */
/*                                                                   */
/*===================================================================*/
void SPU_Enable()
{
  *sp0 = 0xc000; // Enable playing sound
}

/*===================================================================*/
/*                                                                   */
/*                          SPU_Init()                               */
/*                                                                   */
/*===================================================================*/
int SPU_Init()
{
  short status,c,*p;

  *d_pcr       |= 0xb0000;  // enable data transfer through SPU dma channels  
  *spu_mvol_r   = 0x3fff;   // set main volume  
  *spu_mvol_l   = 0x3FFF;
  *sp0          = 0x0;      // reset spu
  SPU_Wait();

  *spu_u_B      = 0x4;
  *spu_u_1      = 0x0;      // set reverberation depth
  *spu_u_2      = 0x0; 
  *spu_keyOFF_1 = 0xffff;   // set all voices off 
  *spu_keyOFF_2 = 0xffff; 
  *spu_u_3      = 0x0; 
  *spu_u_4      = 0x0;
  *spu_u_7      = 0x0; 
  *spu_u_8      = 0x0; 
  *spu_u_9      = 0x0; 
  *spu_u_A      = 0x0;
  *cd_mvol_l    = 0x0;     // set cd main volume 
  *cd_mvol_r    = 0x0;
  *spu_u_5      = 0x0; 
  *spu_u_6      = 0x0; 
  status        = *sp1 & 0x7ff;
  *spu_sb_addr  = 0x200;
  SPU_Wait();

  for ( c = 0; c < 8; ++c )
    *spu_u_D = 0x0707;

  SPU_Wait();

  if ( *sp1 & 0x400 )
  {
    for ( c = 0; *sp1 & 0x400; ++c )
    {
      if ( c > 3841 )
        return -1;
    }
  }
  SPU_Wait();
  SPU_Wait();

  *sp0 = *sp0 & 0xffcf; 
  if ( status != *sp1 & 0x7ff )
  {
    for ( c = 0; status != *sp1 & 0x7ff; ++c )
    {
      if ( c > 3841 )
        return -1;
    }
  }

  // set voices on off
  *spu_keyON_1 = 0xffff;
  *spu_keyON_2 = 0xff;
  SPU_Wait();
  SPU_Wait();
  SPU_Wait();
  SPU_Wait();

  *spu_keyOFF_1 = 0xffff;
  *spu_keyOFF_2 = 0xff;
  SPU_Wait();
  SPU_Wait();
  SPU_Wait();
  SPU_Wait();

  *sp0     = 0x8000;   // set spu on
  *spu_u_E = 0x1fff;
}

/*===================================================================*/
/*                                                                   */
/*                       SPU_UploadData()                            */
/*                                                                   */
/*===================================================================*/
int SPU_UploadData(u_char *addr, u_long sb_addr, u_long size)
{
  long c,i;
  
  *spu_sb_addr = sb_addr >> 3 ;
  
  if ( *spu_sb_addr != sb_addr >> 3 )
  {
    for ( c = 0; *spu_sb_addr != sb_addr >> 3; ++c )
    {
      if ( c > 3841 * 2 ) /* by bero */
        return 0;
    }
  }

  *sp0 = *sp0 & 0xffcf | 0x20 ; // CPU to SPU
  
  if ( *sp0 & 0x30 != 0x20 ) // if spu not ready to get data
  {
    for ( c = 0; *sp0 & 0x30 != 0x20; ++c ) // wait
    {
      if ( c >3841 ) // thats too much
        return 0;
    }
  }

  *spu_delay = *spu_delay & 0xf0ffffff | 0x20000000;
  
  *spu_dma_1 = (u_long)addr;
  *spu_dma_2 = ( ( size + 0x3f ) >> 6 ) << 16 | 0x10;
  *spu_dma_3 = 0x01000201 ;       // start CPU->SPU transfer
  while ( *spu_dma_3 & 0x01000000 )
    ; // wait until transfer finished
  
  return 1;
}

/*===================================================================*/
/*                                                                   */
/*                      SPU_VoiceSetData()                           */
/*                                                                   */
/*===================================================================*/
void SPU_VoiceSetData( u_short voice, u_long sb_addr )
{
  // dont forget that short is here 2 bytes long (not 1 byte)
  volatile short *b = voice_base + ( voice * 8 );
  
  *b       = 16000;
  *(b+0x1) = 16000;
  *(b+0x2) = 1;
  *(b+0x3) = sb_addr >> 3;
  *(b+0x4) = 0x80ff;
  *(b+0x5) = 0x1fff;
  *(b+0x6) = 0;
  *(b+0x7) = 0; // im not sure 
}

/*===================================================================*/
/*                                                                   */
/*                          SPU_VoiceOn()                            */
/*                                                                   */
/*===================================================================*/
void SPU_VoiceOn( u_char voice )
{
  u_long mask;
  
  mask = 1 << voice;
  *spu_keyON_1 = mask & 0xffff; 
  *spu_keyON_2 = mask >> 16; 
  SPU_Wait();
  SPU_Wait();
  SPU_Wait();
  SPU_Wait();
}

/*===================================================================*/
/*                                                                   */
/*                         SPU_VoiceOff()                            */
/*                                                                   */
/*===================================================================*/
void SPU_VoiceOff( u_char voice )
{
  u_long mask;
  
  mask = 1 << voice;
  *spu_keyOFF_1 = mask & 0xffff; 
  *spu_keyOFF_2 = mask >> 16; 
  SPU_Wait();
  SPU_Wait();
  SPU_Wait();
  SPU_Wait();
}

/*===================================================================*/
/*                                                                   */
/*                           SPU_Wait()                              */
/*                                                                   */
/*===================================================================*/
int SPU_Wait (void)
{
  int c;
  long vara = 13;
  
  // wait for some clock cycles
  for ( c = 0; c < 0x60; ++c )
    vara = ( ( vara << 1 ) + vara ) << 2 + vara ; // *13
  
  if ( *sp1 & 0x7ff ) // if spu isnt ready then wait
  {
    for ( c = 0; *sp1 & 0x7ff; ++c )
    {
      if ( c > 3841 ) // thats too long 
        return 0;
    }
  }
}

/*===================================================================*/
/*                                                                   */
/*            SPU_VoiceVol() : Set Voice n Volume (0 -> 0x3fff)      */
/*                                                                   */
/*===================================================================*/
void SPU_VoiceVol( u_short voice, u_short vol_left, u_short vol_right )
{
  volatile short *b = voice_base + ( voice * 8 );
  
  *b = vol_left;
  *( b + 1 ) = vol_right;
}

/*===================================================================*/
/*                                                                   */
/*                         SPU_VoicePitch()                          */
/*                                                                   */
/*===================================================================*/
void SPU_VoicePitch( u_char voice, u_short pitch )
{
  volatile short *b = voice_base + ( voice * 8 );
  
  *( b + 2 ) = pitch;
}
